from kutana.controllers.basic import BasicController
from kutana.plugindata import Message
from kutana.exceptions import ExitException


class DumpingController(BasicController):
    """Shoots target texts once."""

    type = "dumping"

    def __init__(self, *texts):
        self.die = False
        self.queue = list(texts)

    async def async_print(self, *args, **kwargs):
        print(*args, **kwargs)

    async def setup_env(self, update, eenv):
        eenv["reply"] = self.async_print

    @staticmethod
    async def convert_to_message(update, eenv):
        return Message(
            update, (), "U", "KU", update
        )

    async def get_background_coroutines(self, ensure_future):
        return []

    async def get_receiver_coroutine_function(self):
        async def rec():
            if self.die:
                raise ExitException

            self.die = True

            return self.queue

        return rec

    async def dispose(self):
        pass



from kutana.controllers.vk.methodclasses import (
    upload_doc_class, upload_photo_class, reply_concrete_class
)
from kutana.controllers.vk.converter import convert_to_message
from kutana.controllers.vk.vkcontrollerdata import VKRequest, VKResponse
from kutana.controllers.basic import BasicController
from kutana.plugindata import Attachment
from kutana.logger import logger
import asyncio
import aiohttp
import json


class VKController(BasicController):
    """Class for receiving updates from vk.com.
    Controller requires group's token. You can specify settings for
    groups.setLongPollSettings with argument `longpoll_settings`.
    """

    type = "vk"

    def __init__(self, token, longpoll_settings=None):
        if not token:
            raise ValueError("No `token` specified")

        if not longpoll_settings:
            longpoll_settings = {}

        self.session = None
        self.group_id = None
        self.longpoll = None

        self.running = True
        self.requests_queue = []

        self.version = "5.80"
        self.token = token
        self.longpoll_settings = longpoll_settings

        self.api_url = "https://api.vk.com/method/{{}}?access_token={}&v={}" \
            .format(self.token, self.version)
        self.longpoll_url = "{}?act=a_check&key={}&wait=25&ts={}"

    async def raw_request(self, method, **kwargs):
        """Perform api request to vk.com"""

        if not self.session:
            raise RuntimeError("Session is not created yet")

        url = self.api_url.format(method)

        data = {}

        for k, v in kwargs.items():
            if v is not None:
                data[k] = v

        try:
            async with self.session.post(url, data=data) as response:
                raw_respose = await response.json()

        except Exception as e:
            return VKResponse(
                error=True,
                kutana_error=(type(e), str(e)),
                response_error=("", ""),
                response="",
                execute_errors=""
            )

        if raw_respose is None or not isinstance(raw_respose, dict):
            return VKResponse(
                error=True,
                kutana_error=("", ""),
                response_error=("Unknown response", raw_respose),
                response="",
                execute_errors=raw_respose.get("execute_errors", "")
            )

        if not raw_respose.get("response") or raw_respose.get("error"):
            return VKResponse(
                error=True,
                kutana_error=("", ""),
                response_error=("Error", raw_respose["error"]),
                response=raw_respose.get("response", ""),
                execute_errors=raw_respose.get("execute_errors", "")
            )

        return VKResponse(
            error=False,
            kutana_error=("", ""),
            response_error=("", ""),
            response=raw_respose["response"],
            execute_errors=raw_respose.get("execute_errors", "")
        )

    async def request(self, method, **kwargs):
        """Shedule execution of request to vk.com and return result."""

        request = VKRequest(
            method,
            **kwargs
        )

        self.requests_queue.append(request)

        await request

        try:
            return await asyncio.wait_for(request, timeout=10)

        except asyncio.TimeoutError:
            return VKResponse(
                error=True,
                kutana_error=(
                    "Timeout",
                    "Request took too long and was forgotten."
                ),
                response_error=("", ""),
                response="",
                execute_errors=""
            )

    async def send_message(self, message, peer_id, attachment=None,
            sticker_id=None, payload=None, keyboard=None):
        """Send message to target peer_id wiith parameters."""

        if isinstance(attachment, Attachment):
            attachment = [attachment]

        if isinstance(attachment, (list, tuple)):
            new_attachment = ""

            for a in attachment:
                if isinstance(a, Attachment):
                    new_attachment += \
                        "{}{}_{}".format(a.type, a.owner_id, a.id) + \
                        ("_" + a.access_key if a.access_key else "")

                else:
                    new_attachment += str(a)

                new_attachment += ","

            attachment = new_attachment

        return await self.request(
            "messages.send",
            message=message,
            peer_id=peer_id,
            attachment=attachment,
            sticker_id=sticker_id,
            payload=sticker_id,
            keyboard=keyboard
        )\

    async def setup_env(self, update, eenv):
        peer_id = update["object"].get("peer_id")

        if update["type"] == "message_new":
            eenv["reply"] = reply_concrete_class(self, peer_id)

        eenv["send_message"] = self.send_message

        eenv["upload_photo"] = upload_photo_class(self, peer_id)
        eenv["upload_doc"] = upload_doc_class(self, peer_id)

        eenv["request"] = self.request

    async def convert_to_message(self, update, eenv):
        return await convert_to_message(update, eenv)

    @staticmethod
    async def _set_results_to_requests(result, requests):
        err_no = 0

        for res, req in zip(result.response, requests):
            if res is False:
                if len(result.execute_errors) > err_no:
                    known_error = result.execute_errors[err_no]
                    err_no += 1

                else:
                    known_error = ""

                res = VKResponse(
                    error=True,
                    kutana_error=("", ""),
                    response_error=(
                        "Error" if known_error else "Unknown error",
                        known_error
                    ),
                    response="",
                    execute_errors=result.execute_errors
                )

            else:
                res = VKResponse(
                    error=False,
                    kutana_error=("", ""),
                    response_error=("", ""),
                    response=res,
                    execute_errors=""
                )

            try:
                req.set_result(res)
            except asyncio.InvalidStateError:
                pass

    async def _msg_exec_loop(self, ensure_future):
        while self.running:
            requests = []

            await asyncio.sleep(0)

            for _ in range(25):
                if not self.requests_queue:
                    break

                requests.append(self.requests_queue.pop(0))

            if not requests:
                continue

            code = "return ["

            for r in requests:
                code += "API.{}({}),".format(
                    r.method, json.dumps(r.kwargs, ensure_ascii=False)
                )

            code += "];"

            result = await self.raw_request("execute", code=code)

            if result.error:
                logger.error(result.kutana_error + result.response_error)

            await ensure_future(self._set_results_to_requests(result, requests))

    async def get_background_coroutines(self, ensure_future):
        return (self._msg_exec_loop(ensure_future),)

    async def update_longpoll_data(self):
        longpoll = await self.raw_request("groups.getLongPollServer", group_id=self.group_id)

        if longpoll.error:
            raise ValueError(
                "Couldn't get longpoll information\n{}"
                .format(
                    longpoll.kutana_error +
                    longpoll.response_error
                )
            )

        self.longpoll = {
            **longpoll.response
        }

    async def receiver(self):
        async with self.session.post(self.longpoll_url.format(
            self.longpoll["server"],
            self.longpoll["key"],
            self.longpoll["ts"],
        )) as resp:
            try:
                response = await resp.json()
            except Exception:
                return []

        if "ts" in response:
            self.longpoll["ts"] = response["ts"]

        if "failed" in response:
            if response["failed"] in (2, 3, 4):
                await self.update_longpoll_data()

            return

        updates = []

        for update in response["updates"]:
            if "type" not in update or "object" not in update:
                continue

            updates.append(update)

        return updates

    async def get_receiver_coroutine_function(self):
        self.session = aiohttp.ClientSession()

        current_group_s = await self.raw_request("groups.getById")

        if current_group_s.error or not current_group_s.response[0].get("id"):
            raise ValueError(
                "Couldn't get group information\n{}"
                .format(
                    current_group_s.kutana_error +
                    current_group_s.response_error
                )
            )

        self.group_id = current_group_s.response[0]["id"]

        await self.raw_request(
            "groups.setLongPollSettings",
            group_id=self.group_id,
            api_version=self.version,
            enabled=1,
            **{
                **dict(
                    message_reply=0,
                    message_allow=0,
                    message_deny=0,
                    message_edit=0,
                    photo_new=0,
                    audio_new=0,
                    video_new=0,
                    wall_reply_new=0,
                    wall_reply_edit=0,
                    wall_reply_delete=0,
                    wall_reply_restore=0,
                    wall_post_new=0,
                    wall_repost=0,
                    board_post_new=0,
                    board_post_edit=0,
                    board_post_restore=0,
                    board_post_delete=0,
                    photo_comment_new=0,
                    photo_comment_edit=0,
                    photo_comment_delete=0,
                    photo_comment_restore=0,
                    video_comment_new=0,
                    video_comment_edit=0,
                    video_comment_delete=0,
                    video_comment_restore=0,
                    market_comment_new=0,
                    market_comment_edit=0,
                    market_comment_delete=0,
                    market_comment_restore=0,
                    poll_vote_new=0,
                    group_join=0,
                    group_leave=0,
                    group_change_settings=0,
                    group_change_photo=0,
                    group_officers_edit=0,
                    user_block=0,
                    user_unblock=0
                ),
                **self.longpoll_settings
            }
        )

        await self.update_longpoll_data()

        logger.info("logged in as \"{}\" ( https://vk.com/id{} )".format(
            current_group_s.response[0]["name"],
            current_group_s.response[0]["id"]
        ))

        return self.receiver

    async def dispose(self):
        self.running = False

        if self.session:
            await self.session.close()



from collections import namedtuple


Message = namedtuple(
    "Message",
    "text attachments from_id peer_id raw_update"
)

Message.__doc__ = "Text message witch possible attachments."


Attachment = namedtuple(
    "Attachment",
    "type id owner_id access_key link raw_attachment"
)

Attachment.__doc__ = "Detailed information about attachment."



from collections import namedtuple
import asyncio


VKResponse = namedtuple(
    "VKResponse",
    "error kutana_error response_error response execute_errors"
)


class VKRequest(asyncio.Future):
    __slots__ = ("mthod", "kwargs")

    def __init__(self, method, **kwargs):
        super().__init__()

        self.method = method
        self.kwargs = kwargs



from kutana.plugindata import Message, Attachment
import re


def convert_to_attachment(attachment, attachment_type=None):
    if "type" in attachment and attachment["type"] in attachment:
        body = attachment[attachment["type"]]
        attachment_type = attachment["type"]
    else:
        body = attachment

    if "sizes" in body:
        link = body["sizes"][-1]["url"]  # src

    elif "url" in body:
        link = body["url"]

    else:
        link = None

    return Attachment(
        attachment_type,
        body.get("id"),
        body.get("owner_id"),
        body.get("access_key"),
        link,
        attachment
    )


naive_cache = {}

async def resolveScreenName(screen_name, eenv):  # pragma: no cover
    if screen_name in naive_cache:
        return naive_cache[screen_name]

    result = await eenv.request(
        "utils.resolveScreenName",
        screen_name=screen_name
    )

    naive_cache[screen_name] = result

    return result


async def convert_to_message(update, eenv):
    if update["type"] != "message_new":
        return None

    obj = update["object"]

    text = obj["text"]

    if "conversation_message_id" in obj:
        cursor = 0
        new_text = ""

        for m in re.finditer(r"\[(.+?)\|.+?\]", text):
            resp = await resolveScreenName(m.group(1), eenv)

            new_text += text[cursor : m.start()]

            cursor = m.end()

            if resp.error or resp.response["object_id"] == update["group_id"]:
                continue

            new_text += text[m.start() : m.end()]

        new_text += text[cursor :]

        text = new_text.lstrip()

    return Message(
        text,
        tuple(convert_to_attachment(a) for a in obj["attachments"]),
        obj.get("from_id"),
        obj.get("peer_id"),
        update
    )



from kutana.controllers.vk.converter import convert_to_attachment
import aiohttp
import json


async def upload_file_to_vk(ctrl, upload_url, data):
    upload_result_resp = await ctrl.session.post(
        upload_url, data=data
    )

    if not upload_result_resp:
        return None

    upload_result_text = await upload_result_resp.text()

    if not upload_result_text:
        return None

    try:
        upload_result = json.loads(upload_result_text)

        if "error" in upload_result:
            raise Exception

    except Exception:
        return None

    return upload_result


class reply_concrete_class():
    """Class-method for replying to messages."""

    def __init__(self, ctrl, peer_id):
        self.ctrl = ctrl
        self.peer_id = peer_id

    async def __call__(self, message, attachment=None, sticker_id=None, payload=None, keyboard=None):
        return await self.ctrl.send_message(
            message,
            self.peer_id,
            attachment,
            sticker_id,
            payload,
            keyboard
        )


class upload_doc_class():
    """Class-method for uploading documents.

    Pass peer_id=False to upload with docs.getWallUploadServer.
    """

    def __init__(self, ctrl, peer_id):
        self.ctrl = ctrl
        self.peer_id = peer_id

    async def __call__(self, file, peer_id=None, group_id=None,
            doctype="doc", filename=None):
        if filename is None:
            filename = "file.png"

        if peer_id is None:
            peer_id = self.peer_id

        if isinstance(file, str):
            with open(file, "rb") as o:
                file = o.read()

        if peer_id:
            upload_data = await self.ctrl.request(
                "docs.getMessagesUploadServer", peer_id=peer_id, type=doctype
            )

        else:
            upload_data = await self.ctrl.request(
                "docs.getWallUploadServer",
                group_id=group_id or self.ctrl.group_id
            )

        if "upload_url" not in upload_data.response:
            return None

        upload_url = upload_data.response["upload_url"]

        data = aiohttp.FormData()
        data.add_field("file", file, filename=filename)

        upload_result = await upload_file_to_vk(self.ctrl, upload_url, data)

        if not upload_result:
            return None

        attachments = await self.ctrl.request(
            "docs.save", **upload_result
        )

        if not attachments.response:
            return None

        return convert_to_attachment(
            attachments.response[0], "doc"
        )


class upload_photo_class():
    """Class-method for uploading documents."""

    def __init__(self, ctrl, peer_id):
        self.ctrl = ctrl
        self.peer_id = peer_id

    async def __call__(self, file, peer_id=None):
        if peer_id is None:
            peer_id = self.peer_id

        if isinstance(file, str):
            with open(file, "rb") as o:
                file = o.read()

        upload_data = await self.ctrl.request(
            "photos.getMessagesUploadServer", peer_id=peer_id
        )

        if "upload_url" not in upload_data.response:
            return None

        upload_url = upload_data.response["upload_url"]

        data = aiohttp.FormData()
        data.add_field("photo", file, filename="image.png")

        upload_result = await upload_file_to_vk(self.ctrl, upload_url, data)

        if not upload_result:
            return None

        attachments = await self.ctrl.request(
            "photos.saveMessagesPhoto", **upload_result
        )

        if not attachments.response:
            return None

        return convert_to_attachment(
            attachments.response[0], "photo"
        )



from kutana import Executor, logger
from test_framework import KutanaTest
import logging


logging.disable(logging.INFO)


class TestExecutors(KutanaTest):
    def tearDown(self):
        logger.setLevel(logging.ERROR)

    def test_just_dumping(self):
        self.target = ["message"] * 5

        with self.dumping_controller(self.target):

            async def new_update(update, eenv):
                if eenv.ctrl_type == "dumping":
                    self.actual.append(update)

                else:
                    self.assertEqual(eenv.ctrl_type, "kutana")

            self.kutana.executor.register(new_update)

    def test_exception(self):
        self.called = 0

        with self.dumping_controller(["message"] * 5):

            async def new_update(update, eenv):
                if eenv.ctrl_type == "dumping":
                    raise Exception

            self.kutana.executor.register(new_update)

            async def new_error(update, eenv):
                self.assertTrue(eenv.exception)

                self.called += 1

                return "DONE"

            async def new_error_no(update, eenv):
                self.assertTrue(False)

            self.kutana.executor.register(new_error, error=True)
            self.kutana.executor.register(new_error_no, error=True)

            logger.setLevel(logging.CRITICAL)

        self.assertEqual(self.called, 5)

    def test_default_exception_handle(self):
        self.called = 0

        async def my_faked_reply(mes):
            self.assertEqual(mes, "Произошла ошибка! Приносим свои извинения.")
            self.called += 1

        with self.dumping_controller(["message"] * 5):

            async def new_update(update, eenv):
                if eenv.ctrl_type == "dumping":
                    eenv.reply = my_faked_reply
                    raise Exception

            self.kutana.executor.register(new_update)

            logger.setLevel(logging.CRITICAL)

        self.assertEqual(self.called, 5)


    def test_two_dumping(self):
        self.target = ["message"] * 5

        with self.dumping_controller(self.target):
            with self.dumping_controller(self.target):
                self.target *= 2

                async def new_update(update, eenv):
                    if eenv.ctrl_type == "dumping":
                        self.actual.append(update)

                    else:
                        self.assertEqual(eenv.ctrl_type, "kutana")

                self.kutana.executor.register(new_update)

    def test_two_callbacks_and_two_dumpings(self):
            self.target = ["message"] * 5

            with self.dumping_controller(self.target):
                with self.dumping_controller(self.target):
                    self.target *= 4

                    async def new_update_1(update, eenv):
                        if eenv.ctrl_type == "dumping":
                            self.actual.append(update)

                    async def new_update_2(update, eenv):
                        if eenv.ctrl_type == "dumping":
                            self.actual.append(update)

                    self.kutana.executor.register(new_update_1)
                    self.kutana.executor.register(new_update_2)

    def test_decorate_or_call(self):
        self.target = ["message"] * 5

        with self.dumping_controller(self.target):
            self.target *= 2

            @self.kutana.executor.register()
            async def new_update(update, eenv):
                if eenv.ctrl_type == "dumping":
                    self.actual.append(update)

            self.kutana.executor.register(new_update)

    def test_merge_executors(self):
        exec1 = Executor()
        exec2 = Executor()

        async def cor1(*args, **kwargs):
            pass

        async def cor2(*args, **kwargs):
            pass

        exec1.register(cor1)
        exec2.register(cor2)

        exec1.callbacks_owners.append(1)
        exec2.callbacks_owners.append(2)

        exec1.add_callbacks_from(exec2)

        self.assertEqual(exec1.callbacks, [cor1, cor2])
        self.assertEqual(exec1.error_callbacks, [])
        self.assertEqual(exec1.callbacks_owners, [1, 2])



import collections


class objdict(collections.MutableMapping):
    """dict with gettings and settings throught dot."""

    __slots__ = ("_store",)

    def __init__(self, *args, **kwargs):
        self._store = dict()
        self.update(dict(*args, **kwargs))

    def __getattr__(self, name):
        return self.__getitem__(name)

    def __setattr__(self, name, value):
        if name in self.__slots__:
            return object.__setattr__(self, name, value)

        return self.__setitem__(name, value)

    def __getitem__(self, key):
        return self._store[key]

    def __setitem__(self, key, value):
        if isinstance(value, (dict, objdict)):
            self._store[key] = objdict(value)
        else:
            self._store[key] = value

    def __delitem__(self, key):
        del self._store[key]

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __str__(self):
        return str(self._store)

    def __repr__(self):  # pragma: no cover
        return repr(self._store)


class icedict(collections.Mapping):
    """Immutable dict.
    Thanks
    https://stackoverflow.com/questions/2703599/what-would-a-frozen-dict-be
    """

    __slots__ = ("_store",)

    def __init__(self, *args, **kwargs):
        self._store = dict(*args, **kwargs)

    def __eq__(self, other):
        if isinstance(other, icedict):
            return self._store == other._store

        return self._store == other

    def __iter__(self):
        return iter(self._store)

    def __len__(self):
        return len(self._store)

    def __getattr__(self, name):
        return self.__getitem__(name)

    def __getitem__(self, key):
        return self._store[key]

    def __str__(self):
        return str(self._store)

    def __repr__(self):  # pragma: no cover
        return repr(self._store)



class BasicController:  # pragma: no cover
    type = "basic"

    async def setup_env(self, update, eenv):
        """Installs wanted methods and values for update into eenv."""

        raise NotImplementedError

    async def convert_to_message(self, update, eenv):
        """Converts raw update to instances of :class:`.Message`
        and :class:`.Attachmnet` for plugins.
        """

        raise NotImplementedError

    async def get_background_coroutines(self, ensure_future):
        """Return list of tasks to be executed in background in Kutana."""

        raise NotImplementedError

    async def get_receiver_coroutine_function(self):
        """Return coroutine function for getting updates from vk.com."""

        raise NotImplementedError

    async def dispose(self):
        """Free used resources."""

        raise NotImplementedError



from kutana.logger import logger
import importlib.util
import json
import os


def load_configuration(target, path):
    with open(path, "r") as fh:
        config = json.load(fh)

    return config.get(target)


def import_plugin(name, path):
    """Import plugin from specified path with specified name."""

    spec = importlib.util.spec_from_file_location(name, path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)

    return module.plugin

def load_plugins(plugins_folder, verbose=True):
    """Import all plugins from target folder recursively."""

    found_plugins = []

    for pack in os.walk(plugins_folder):
        for filename in pack[2]:
            if "_" == filename[:1] or ".py" != filename[-3:]:
                continue

            path_to_module = os.path.join(plugins_folder, filename)

            if verbose:
                logger.info("Loading plugin \"{}\"".format(path_to_module))

            plugin = import_plugin(path_to_module, path_to_module)

            found_plugins.append(plugin)

        for folder in pack[1]:
            found_plugins += load_plugins(os.path.join(plugins_folder, folder))

    return sorted(found_plugins, key=lambda pl: pl.order)



from kutana.tools.structures import objdict
import shlex
import re


class Plugin():
    """Class for craeting extensions for kutana engine."""

    def __init__(self, **kwargs):
        self._callbacks = []
        self._callbacks_raw = []
        self._callbacks_dispose = []
        self._callback_startup = None

        self.order = 50

        for k, v in kwargs.items():
            setattr(self, k, v)

    @staticmethod
    def done_if_none(value):
        """Return "DONE" if value is None. Otherwise return value."""

        if value is None:
            return "DONE"

        return value

    async def proc_update(self, update, eenv):
        """Method for processing updates."""

        if eenv.ctrl_type == "kutana":
            if update["update_type"] == "startup":
                if self._callback_startup:
                    await self._callback_startup(update["kutana"], update)

            return

        env = objdict(eenv=eenv, **eenv)

        if eenv.get("_cached_message"):
            message = eenv["_cached_message"]

        else:
            message = await eenv.convert_to_message(update, eenv)

            eenv["_cached_message"] = message

        if message is None:
            if not self._callbacks_raw:
                return

            arguments = {
                "env": env,
                "update": update
            }

            callbacks = self._callbacks_raw

        else:
            arguments = {
                "env": env,
                "message": message,
                "attachments": message.attachments
            }

            callbacks = self._callbacks

        for callback in callbacks:
            comm = await callback(**arguments)

            if comm == "DONE":
                return "DONE"

    async def dispose(self):
        """Free resources and prepare for shutdown."""

        for callback in self._callbacks_dispose:
            await callback()

    def add_callbacks(self, *callbacks):
        """Add callbacks for processing updates."""

        for callback in callbacks:
            self._callbacks.append(callback)

    def on_dispose(self):
        """Returns decorator for adding callbacks which is triggered when
        everything is going to shutdown.
        """

        def decorator(coro):
            self._callbacks_dispose.append(coro)

            return coro

        return decorator

    def on_startup(self):
        """Returns decorator for adding callbacks which is triggered
        at the startup of kutana. Decorated coroutine receives kutana
        object and some information in update.
        """

        def decorator(coro):
            self._callback_startup = coro

            return coro

        return decorator

    def on_raw(self):
        """Returns decorator for adding callbacks which is triggered
        every time when update can't be turned into `Message` or
        `Attachment` object. Arguments `env` and raw `update`
        is passed to callback.
        """

        def decorator(coro):
            self._callbacks_raw.append(coro)

            return coro

        return decorator

    def on_text(self, *texts):
        """Returns decorator for adding callbacks which is triggered
        when the message and any of the specified text are fully matched.
        """

        def decorator(coro):
            check_texts = list(text.strip().lower() for text in texts)

            async def wrapper(*args, **kwargs):
                if kwargs["message"].text.strip().lower() in check_texts:
                    comm = self.done_if_none(await coro(*args, **kwargs))

                    if comm == "DONE":
                        return "DONE"

            self.add_callbacks(wrapper)

            return wrapper

        return decorator

    def on_has_text(self, *texts):
        """Returns decorator for adding callbacks which is triggered
        when the message contains any of the specified texts.

        Fills env for callback with:

        - "found_text" - text found in message.
        """

        def decorator(coro):
            check_texts = tuple(text.strip().lower() for text in texts) or ("",)

            async def wrapper(*args, **kwargs):
                check_text = kwargs["message"].text.strip().lower()

                for text in check_texts:
                    if text not in check_text:
                        continue

                    kwargs["env"]["found_text"] = text

                    comm = self.done_if_none(await coro(*args, **kwargs))

                    if comm == "DONE":
                        return "DONE"

            self.add_callbacks(wrapper)

            return wrapper

        return decorator

    def on_startswith_text(self, *texts):
        """Returns decorator for adding callbacks which is triggered
        when the message starts with any of the specified texts.

        Fills env for callback with:

        - "body" - text without prefix.
        - "args" - text without prefix splitted in bash-like style.
        - "prefix" - prefix.
        """

        def decorator(coro):
            check_texts = tuple(text.lstrip().lower() for text in texts)

            def search_prefix(message):
                for text in check_texts:
                    if message.startswith(text):
                        return text

                return None

            async def wrapper(*args, **kwargs):
                search_result = search_prefix(kwargs["message"].text.lower())

                if search_result is None:
                    return

                kwargs["env"]["body"] = kwargs["message"].text[len(search_result):].strip()
                kwargs["env"]["args"] = shlex.split(kwargs["env"]["body"])
                kwargs["env"]["prefix"] = kwargs["message"].text[:len(search_result)].strip()

                comm = self.done_if_none(await coro(*args, **kwargs))

                if comm == "DONE":
                    return "DONE"

            self.add_callbacks(wrapper)

            return wrapper

        return decorator

    def on_regexp_text(self, regexp, flags=0):
        """Returns decorator for adding callbacks which is triggered
        when the message matches the specified regular expression.

        Fills env for callback with:

        - "match" - match.
        """

        if isinstance(regexp, str):
            compiled = re.compile(regexp, flags=flags)

        else:
            compiled = regexp

        def decorator(coro):
            async def wrapper(*args, **kwargs):
                match = compiled.match(kwargs["message"].text)

                if not match:
                    return

                kwargs["env"]["match"] = match

                comm = self.done_if_none(await coro(*args, **kwargs))

                if comm == "DONE":
                    return "DONE"

            self.add_callbacks(wrapper)

            return wrapper

        return decorator

    def on_attachment(self, *types):
        """Returns decorator for adding callbacks which is triggered
        when the message has attachments of the specified type
        (if no types specified, then any attachments).
        """

        def decorator(coro):
            async def wrapper(*args, **kwargs):
                if not kwargs["attachments"]:
                    return

                if types:
                    for a in kwargs["attachments"]:
                        if a.type in types:
                            break
                    else:
                        return

                comm = self.done_if_none(await coro(*args, **kwargs))

                if comm == "DONE":
                    return "DONE"

            self.add_callbacks(wrapper)

            return wrapper

        return decorator



from kutana.logger import logger
import sys


class Executor:
    """Class performing update processing."""

    def __init__(self):
        self.callbacks = []
        self.error_callbacks = []

        self.callbacks_owners = []

    def add_callbacks_from(self, executor):
        """Updates with callbacks from other executor."""

        self.callbacks += executor.callbacks
        self.error_callbacks += executor.error_callbacks

        self.callbacks_owners += executor.callbacks_owners

    def register_plugins(self, *plugins):
        """Register callbacks from plugins."""

        for plugin in plugins:
            if hasattr(plugin, "proc_update"):
                self.register(plugin.proc_update)

            if hasattr(plugin, "proc_error"):  # pragma: no cover
                self.register(plugin.proc_error, error=True)

    def register(self, *callbacks, error=False):
        """Register callbacks."""

        def _register(coroutine):
            if error:
                self.error_callbacks.append(coroutine)

            else:
                self.callbacks.append(coroutine)

            return coroutine

        for callback in callbacks:
            if hasattr(callback, "__self__"):
                self.callbacks_owners.append(callback.__self__)

            _register(callback)

        return _register

    async def __call__(self, update, eenv):
        """Process update from controller."""

        try:
            for cb in self.callbacks:
                comm = await cb(update, eenv)

                if comm == "DONE":
                    break

        except Exception as e:
            logger.exception(
                "\"{}::{}\"on update {} from {}".format(
                    sys.exc_info()[0].__name__, e, update, eenv.ctrl_type
                )
            )

            eenv["exception"] = e

            if not self.error_callbacks:
                if "reply" in eenv:
                    return await eenv.reply("Произошла ошибка! Приносим свои извинения.")

            for cb in self.error_callbacks:
                comm = await cb(update, eenv)

                if comm == "DONE":
                    break

    async def dispose(self):
        """Free resources and prepare for shutdown."""

        for callback_owner in self.callbacks_owners:
            if hasattr(callback_owner, "dispose"):
                await callback_owner.dispose()



from kutana import Plugin
from test_framework import KutanaTest
import re


class TestPlugins(KutanaTest):
    def test_plugin_creation(self):
        plugin = Plugin(name="Name", cmds=["cmd1", "cmd2"])

        self.assertEqual(plugin.name, "Name")  # pylint: disable=E1101
        self.assertEqual(plugin.cmds, ["cmd1", "cmd2"])  # pylint: disable=E1101

        plugin = Plugin(order=5)

    def test_echo_plugin(self):
        queue = ["message", "echo message", "echonotecho"] * 5

        self.target = ["message"] * 5

        with self.dumping_controller(queue) as plugin:
            async def on_echo(message, env, **kwargs):
                self.actual.append(env.body)

            plugin.on_startswith_text("echo ", "echo\n")(on_echo)

    def test_plugin_on_startup(self):
        self.called = 0

        with self.dumping_controller("message") as plugin:
            async def on_startup(message, env, **kwargs):
                self.called += 1

            plugin.on_startup()(on_startup)

        self.assertEqual(self.called, 1)

    def test_args_on_startswith_text(self):
        queue = ["pr a b c", "pr a \"b c\"", "pr a \"\\\"\" b c"]

        queue_answer = [
            ["a", "b", "c"], ["a", "b c"], ["a", "\"", "b", "c"]
        ]

        with self.dumping_controller(queue) as plugin:
            async def on_startswith_text(message, env, **kwargs):
                self.assertEqual(env.args, queue_answer.pop(0))

            plugin.on_startswith_text("pr")(on_startswith_text)

        self.assertFalse(queue_answer)

    def test_plugins_callbacks_done(self):
        self.counter = 0

        with self.dumping_controller(["123"] * 5) as plugin:

            async def on_has_text(message, **kwargs):
                self.counter += 1

                return "DONE"

            plugin.on_has_text("123")(on_has_text)

            async def on_has_text_2(message, **kwargs):
                self.counter += 1

            plugin.on_has_text("123")(on_has_text_2)

        self.assertEqual(self.counter, 5)

    def test_plugins_callbacks_not_done(self):
        self.counter = 0

        with self.dumping_controller(["123"] * 5) as plugin:

            async def on_has_text(message, **kwargs):
                self.counter += 1

                return "GOON"

            plugin.on_has_text("123")(on_has_text)

            async def on_has_text_2(message, **kwargs):
                self.counter += 1

            plugin.on_has_text("123")(on_has_text_2)

        self.assertEqual(self.counter, 10)

    def test_multiple_plugins(self):
        self.counter = 0

        with self.dumping_controller(["msg"] * 2):
            self.plugins.append(Plugin())
            self.plugins.append(Plugin())

            async def on_has_text(message, env, **kwargs):
                self.counter += 1

                self.assertNotIn("key", env)
                env["key"] = "value"

                return "GOON"

            for pl in self.plugins:
                pl.on_has_text()(on_has_text)

        self.assertEqual(self.counter, 6)

    def test_plugin_callbacks(self):
        self.disposed = 0
        self.counter = 0

        with self.dumping_controller(["123"] * 5 + ["321"]) as plugin:
            async def on_123(message, **kwargs):
                self.assertEqual(message.text, "123")
                self.counter += 1

            plugin.on_text("123")(on_123)


            async def on_everything(message, **kwargs):
                self.assertNotEqual(message.text, "123")
                self.counter += 1

            plugin.on_has_text()(on_everything)


            async def on_raw(message, **kwargs):
                raise RuntimeError

            plugin.on_raw()(on_raw)


            async def on_dispose():
                self.disposed += 1

            plugin.on_dispose()(on_dispose)

        self.assertEqual(self.counter, 6)
        self.assertEqual(self.disposed, 1)

    def test_environment_reply(self):
        self.target = ["echo 123"]

        with self.dumping_controller(self.target) as plugin:
            async def on_echo(message, env, **kwargs):
                self.assertIsNotNone(env.reply)
                self.actual.append(message.text)

            plugin.on_startswith_text("echo")(on_echo)

    def test_plugin_onstar(self):
        queue = ["привет", "отлично привет", "ecHo", "ab", "ae", "hello"]
        self.result = 0

        with self.dumping_controller(queue) as plugin:
            async def no_trigger(message, **kwargs):
                self.assertTrue(False)

            plugin.on_attachment("photo", "audio")(no_trigger)


            async def zero_trigger(message, **kwargs):
                self.result |= 1 << 0

            plugin.on_has_text("привет")(zero_trigger)


            async def one_trigger(message, **kwargs):
                self.result |= 1 << 1

            plugin.on_regexp_text(r"a(b|c)")(one_trigger)


            async def two_trigger(message, **kwargs):
                self.result |= 1 << 2

            plugin.on_regexp_text(re.compile(r"a(e|z)"))(two_trigger)


            async def three_trigger(message, **kwargs):
                self.result |= 1 << 3

            plugin.on_startswith_text("hell")(three_trigger)


            async def four_trigger(message, env, **kwargs):
                self.result |= 1 << 4

            plugin.on_text("ecHo")(four_trigger)

        self.assertEqual(self.result, (1 << 5) - 1)



import logging


formatter = logging.Formatter('%(asctime)s [ %(levelname)s ] %(message)s')
level = logging.DEBUG

handler_file = logging.FileHandler('kutana.log')
handler_file.setLevel(level)
handler_file.setFormatter(formatter)

handler_stream = logging.StreamHandler()
handler_stream.setLevel(level)
handler_stream.setFormatter(formatter)

logger = logging.getLogger("kutana")
logger.setLevel(level)
logger.addHandler(handler_stream)
logger.addHandler(handler_file)



from kutana import VKResponse, DumpingController, Executor, load_plugins, \
    objdict, icedict, load_configuration
import kutana.controllers.vk.converter as vk_converter
import unittest
import asyncio


class TestMiscellaneous(unittest.TestCase):
    def test_objdict(self):
        a = objdict()

        self.assertEqual(str(a), "{}")
        self.assertEqual(len(a), 0)
        self.assertEqual(list(i for i in a), [])

        a.a = 10  # pylint: disable=E0237
        self.assertEqual(a.a, 10)
        self.assertEqual(a["a"], 10)
        self.assertEqual(a.get("a"), 10)

        a["b"] = {"a": 5}
        self.assertEqual(a.b.a, 5)

        a["c"] = [1, 2]
        self.assertIs(a.c, a["c"])

        del a["a"]
        del a["b"]
        del a["c"]

        self.assertEqual(a, {})

    def test_icedict(self):
        a = icedict(a=10)

        self.assertEqual(str(a), "{'a': 10}")
        self.assertEqual(len(a), 1)
        self.assertEqual(list(i for i in a), ["a"])
        self.assertEqual(a["a"], 10)
        self.assertEqual(a.a, 10)

        b = icedict(a=10)

        self.assertEqual(a, b)
        self.assertEqual(a, {"a": 10})

        with self.assertRaises(TypeError):
            a["a"] = 10

    def test_functions(self):
        loaded_plugins = load_plugins("test/test_plugins/")

        self.assertEqual(len(loaded_plugins), 1)
        self.assertEqual(loaded_plugins[0].name, "Memory")

        executor = Executor()
        executor.register_plugins(*loaded_plugins)

        loop = asyncio.get_event_loop()

        loop.run_until_complete(
            executor(
                "message", objdict(
                    ctrl_type="dumping",
                    convert_to_message=DumpingController.convert_to_message
                )
            )
        )

        self.assertEqual(loaded_plugins[0].memory, "message")

    def test_load_configuration(self):
        value = load_configuration("key", "test/test_assets/sample.json")

        self.assertEqual(value, "value")

        value = load_configuration("key2", "test/test_assets/sample.json")

        self.assertEqual(value, {"keynkey": "hvalue"})

    def test_vk_conversation(self):
        async def fake_resolveScreenName(*args, **kwargs):
            return VKResponse(False, "", "", {"object_id": 1}, "")

        resolveScreenName = vk_converter.resolveScreenName
        vk_converter.resolveScreenName = fake_resolveScreenName

        loop = asyncio.get_event_loop()

        message = loop.run_until_complete(
            vk_converter.convert_to_message(
                {"object": {"date": 1, "random_id": 0, "fwd_messages": [],
                "important": False, "peer_id": 1,
                "text": "echo [club1|\u0421\u043e] 123", "attachments": [],
                "conversation_message_id": 1411, "out": 0, "from_id": 1,
                "id": 0, "is_hidden": False}, "group_id": 1,
                "type": "message_new"},
                {w: 1 for w in ("reply", "send_msg", "request",
                "upload_photo", "upload_doc")}
            )
        )

        self.assertEqual(message.text, "echo  123")
        self.assertEqual(message.attachments, ())

        message = loop.run_until_complete(
            vk_converter.convert_to_message(
                {"object": {"date": 1, "random_id": 0, "fwd_messages": [],
                "important": False, "peer_id": 1,
                "text": "echo [club1|\u0421\u043e] 123", "attachments": [],
                "conversation_message_id": 1411, "out": 0, "from_id": 1,
                "id": 0, "is_hidden": False}, "group_id": 2,
                "type": "message_new"},
                {w: 1 for w in ("reply", "send_msg", "request",
                "upload_photo", "upload_doc")}
            )
        )

        self.assertEqual(message.text, "echo [club1|\u0421\u043e] 123")
        self.assertEqual(message.attachments, ())

        vk_converter.resolveScreenName = resolveScreenName


if __name__ == '__main__':
    unittest.main()



from kutana import Kutana, VKController, ExitException, Plugin, \
    logger
import unittest
import requests
import logging
import time
import json
import os


logging.disable(logging.INFO)


# These tests requires internet and token of VK user account as well as VK group.
# You can set these values through environment variables:
# TEST_TOKEN and TEST_UTOKEN.
#
# Alternatively you can use json file `configuration_test.json` with format like this:
# {
#   "token": "токен группы",
#   "utoken": "токен пользователя, который может писать в группу"
# }


class TestControllerVk(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        try:
            with open("configuration_test.json") as o:
                cls.conf = json.load(o)

        except FileNotFoundError:
            cls.conf = {
                "vk_token": os.environ.get("TEST_TOKEN", ""),
                "vk_utoken": os.environ.get("TEST_UTOKEN", ""),
            }

        cls.messages_to_delete = set()

        if not cls.conf["vk_token"] or not cls.conf["vk_utoken"]:
            raise unittest.SkipTest("No authorization found for this tests.")

        async def get_receiver_coroutine_function(self):
            actual_reci = await self.original_get_receiver_coroutine_function()

            empty_if_done = [1]

            async def reci():
                if not empty_if_done:
                    raise ExitException

                empty_if_done.pop(0)

                cls.ureq(
                    "messages.setActivity",
                    type="typing",
                    peer_id=-self.group_id
                )

                cls.messages_to_delete.add(
                    str(
                        cls.ureq(
                            "messages.send",
                            message="echo message",
                            peer_id=-self.group_id,
                            attachment= "photo-164328508_456239017," * 2
                        )
                    )
                )

                return await actual_reci()

            return reci

        VKController.original_get_receiver_coroutine_function = VKController.get_receiver_coroutine_function
        VKController.get_receiver_coroutine_function = get_receiver_coroutine_function

        cls.kutana = Kutana()

        cls.kutana.add_controller(
            VKController(
                token=cls.conf["vk_token"],
                longpoll_settings={"message_typing_state": 1}
            )
        )

    @classmethod
    def ureq(cls, method, **kwargs):
        data = {"v": "5.80"}

        for k, v in kwargs.items():
            if v is not None:
                data[k] = v

        response = requests.post(
            "https://api.vk.com/method/{}?access_token={}".format(
                method,
                cls.conf["vk_utoken"]
            ),
            data=data
        ).json()

        time.sleep(0.34)

        return response["response"]

    @classmethod
    def tearDownClass(cls):
        VKController.get_receiver_coroutine_function = VKController.original_get_receiver_coroutine_function

    def tearDown(self):
        if self.messages_to_delete:
            self.ureq("messages.delete", message_ids=",".join(self.messages_to_delete), delete_for_all=1)

            self.messages_to_delete.clear()

    def test_exceptions(self):
        with self.assertRaises(ValueError):
            VKController("")

        with self.assertRaises(RuntimeError):
            self.kutana.loop.run_until_complete(
                VKController("token").raw_request("any.method")
            )

    def test_vk_full(self):
        plugin = Plugin()

        self.called = False
        self.called_on_attachment = False

        async def on_attachment(*args, **kwargs):
            self.called_on_attachment = True
            return "GOON"

        plugin.on_attachment("photo")(on_attachment)

        async def on_regexp(message, attachments, env, **kwargs):
            # Test receiving
            self.assertEqual(env.match.group(1), "message")
            self.assertEqual(env.match.group(0), "echo message")

            self.assertEqual(message.attachments, attachments)
            self.assertEqual(len(attachments), 2)

            self.assertTrue(attachments[0].link)
            self.assertTrue(attachments[1].link)

            # Test sending
            a_image = await env.upload_photo("test/test_assets/author.png")
            a_image = await env.upload_photo("test/test_assets/author.png", peer_id=False)
            a_audio = await env.upload_doc("test/test_assets/girl.ogg", doctype="audio_message", filename="file.ogg")

            self.assertTrue(a_image.id)
            self.assertTrue(a_audio.id)

            resp = await env.reply("Спасибо.", attachment=a_image)

            self.assertTrue(resp.response)

            resp = await env.request("messages.delete", message_ids=str(resp.response), delete_for_all=1)

            self.assertTrue(resp.response)

            # Test failed request
            resp = await env.request("wrong.method")

            self.assertTrue(resp.error)
            self.assertFalse(resp.response)

            self.called = True

        plugin.on_regexp_text(r"echo (.+)")(on_regexp)

        async def on_raw(*args, **kwargs):
            return "GOON"

        plugin.on_raw()(on_raw)

        self.kutana.executor.register_plugins(plugin)

        self.kutana.run()

        self.assertTrue(self.called)
        self.assertTrue(self.called_on_attachment)



from kutana.executor import Executor
from kutana.exceptions import ExitException
from kutana.tools.structures import objdict
import asyncio


class Kutana:
    """Main class for constructing engine."""

    def __init__(self, executor=None, loop=None):
        self.controllers = []
        self.executor = executor or Executor()

        self.loop = loop or asyncio.get_event_loop()

        self.running = True
        self.loops = []
        self.tasks = []
        self.gathered_loops = None

    def add_controller(self, controller):
        """Adding controller to engine."""

        self.controllers.append(controller)

    async def process_update(self, ctrl, update):
        """Prepare environment and process update from controller."""

        eenv = objdict(
            ctrl_type=ctrl.type,
            convert_to_message=ctrl.convert_to_message
        )

        await ctrl.setup_env(update, eenv)

        await self.executor(update, eenv)

    async def ensure_future(self, awaitable):
        """Shurtcut for asyncio.ensure_loop with curretn loop."""

        await asyncio.ensure_future(awaitable, loop=self.loop)

    async def loop_for_controller(self, ctrl):
        """Receive and process updated from target controller."""

        receiver = await ctrl.get_receiver_coroutine_function()

        while self.running:
            for update in await receiver():
                await self.ensure_future(
                    self.process_update(
                        ctrl, update
                    )
                )

            await asyncio.sleep(0.051)

    def run(self):
        """Start engine."""

        asyncio.set_event_loop(self.loop)

        self.loops = []

        for controller in self.controllers:
            self.loops.append(self.loop_for_controller(controller))

            awaitables = self.loop.run_until_complete(
                controller.get_background_coroutines(self.ensure_future)
            )

            for awaitable in awaitables:
                self.loops.append(awaitable)

        self.loop.run_until_complete(
            self.executor(
                {
                    "kutana": self, "update_type": "startup",
                    "callbacks_owners": self.executor.callbacks_owners
                },
                objdict(ctrl_type="kutana")
            )
        )

        try:
            self.gathered = asyncio.gather(*self.loops, *self.tasks)

            self.loop.run_until_complete(self.gathered)
        except (KeyboardInterrupt, ExitException):
            pass

        self.loop.run_until_complete(self.dispose())

    async def dispose(self):
        """Free resources and prepare for shutdown."""

        await self.executor.dispose()

        for controller in self.controllers:
            await controller.dispose()
